#
# File  : Install.py
#
# Copyright (c) 2005, Pilot Systems (http://www.pilotsystems.net)
#
# GNU General Public Licence (GPL)
# 
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA  02111-1307  USA
#
"""Install and uninstall methods

$Id: Install.py 23871 2006-05-24 10:34:33Z dbaty $
"""

from StringIO import StringIO

from Products.CMFCore.utils import getToolByName
from Products.CMFDynamicViewFTI.migrate import migrateFTIs

from Products.StagingAddOn.config import PROJECT_NAME, GLOBALS, CONFIGLETS, PORTLETS, SKIN_DIRECTORIES
from Products.StagingAddOn.content.staging_tool import META_TYPE, ID
from Products.StagingAddOn.content.staging_tool import StagingTool

from Products.Archetypes.public import listTypes, process_types
from Products.Archetypes.Extensions.utils import installTypes, install_subskin


def extendLinesProperty(obj, property, values):
    """Prepend ``property`` of ``obj``  with ``values``

    ``property`` should be the name of a "lines" property. ``values``
    should be a sequence.
    """
    current = list(obj.getProperty(property))
    index = 0
    for value in values:
        if value not in current:
            index += 1
            current.insert(index, value)
    obj.manage_changeProperties(**{property : current})


def removeFromLinesProperty(obj, property, values):
    """Remove each value of `values`` from ``property`` of ``obj``.

    ``property should be the name of a "lines" property. ``values``
    should be a sequence.
    """
    current = list(obj.getProperty(property))
    for value in values:
        if value in current:
            current.remove(value)
    obj.manage_changeProperties(**{property : current})


def install(self):
    """Install this product"""
    out = StringIO()

    ## Install skin
    install_subskin(self, out, GLOBALS)

    ## Add content types
    types = listTypes(PROJECT_NAME)
    installTypes(self, out, types, PROJECT_NAME)

    ## Migrate FTIs
    migrated = migrateFTIs(self, PROJECT_NAME)

    ## Enable portal_factory for the new content types
    factory = getToolByName(self, 'portal_factory')
    new_types, useless1, useless2 = process_types(types, PROJECT_NAME)
    types = factory.getFactoryTypes().keys()
    for new_type in new_types:
        if new_type.portal_type not in types:
            types.append(new_type.portal_type)
    factory.manage_setPortalFactoryTypes(listOfTypeIds=types)

    ## Add tool
    portal = getToolByName(self, 'portal_url').getPortalObject()
    if getToolByName(portal, ID, None) is None:
        ## Temporarily ``global_allow`` the tool
        ttool = getToolByName(portal, 'portal_types')
        type_info = ttool.getTypeInfo(META_TYPE)
        type_info.manage_changeProperties(global_allow=True)
        portal.invokeFactory(META_TYPE, ID)
        type_info.manage_changeProperties(global_allow=False)

    ## Add portlets
    extendLinesProperty(portal, 'right_slots', PORTLETS)

    ## Add configlet
    panel = getToolByName(self, 'portal_controlpanel')
    if panel is not None:
        existing = panel.enumConfiglets(group='Products')
        existing = [c['id'] for c in existing]
        for configlet in CONFIGLETS:
            if configlet['id'] not in existing:
                panel.registerConfiglet(**configlet)

    ## Remove new types from the navigation portlet
    types = listTypes(PROJECT_NAME)
    new_types, useless1, useless2 = process_types(types, PROJECT_NAME)
    new_types = [t.portal_type for t in new_types]
    portal_properties = getToolByName(portal, 'portal_properties')
    nav_props = portal_properties.navtree_properties
    extendLinesProperty(nav_props, 'metaTypesNotToList', new_types)

    ## New types should not appear in the search form
    site_properties = portal_properties.site_properties
    site_properties.manage_changeProperties(types_not_searched=new_types)

    print >> out, "Successfully installed %s." % PROJECT_NAME
    return out.getvalue()


def uninstall(self):
    """Uninstall this product"""
    out = StringIO()
    portal = getToolByName(self, 'portal_url').getPortalObject()

    new_types, _, _ = process_types(listTypes(PROJECT_NAME),
                                    PROJECT_NAME)

    ## Remove tool
    portal.manage_delObjects([ID])

    ## Remove our content types from ``portal_factory``
    factory = getToolByName(portal, 'portal_factory')
    types = factory.getFactoryTypes().keys()
    for new_type in new_types:
        if new_type.portal_type in types:
            types.remove(new_type.portal_type)
    factory.manage_setPortalFactoryTypes(listOfTypeIds=types)

    ## Remove portlets
    removeFromLinesProperty(portal, 'right_slots', PORTLETS)

    ## Remove configlets
    panel = getToolByName(portal, 'portal_controlpanel')
    if panel is not None:
        for configlet in CONFIGLETS:
            panel.unregisterConfiglet(configlet['id'])

    ## Remove layers from the skins
    skins_tool = getToolByName(portal, 'portal_skins')
    selections = skins_tool._getSelections()
    for skin, layers in selections.items():
        layers = layers.split(',')
        for layer in SKIN_DIRECTORIES:
            if layer in layers:
                layers.remove(layer)
        layers = ','.join(layers)
        selections[skin] = layers

    ## Update ``metaTypesNotToList``
    types = listTypes(PROJECT_NAME)
    new_types, useless1, useless2 = process_types(types, PROJECT_NAME)
    new_types = [t.portal_type for t in new_types]
    portal_properties = getToolByName(portal, 'portal_properties')
    nav_props = portal_properties.navtree_properties
    removeFromLinesProperty(nav_props, 'metaTypesNotToList', new_types)

    ## Update ``types_not_searched``
    site_properties = portal_properties.site_properties
    removeFromLinesProperty(site_properties, 'types_not_searched',
                            new_types)

    print >> out, "Successfully uninstalled %s." % PROJECT_NAME
    return out.getvalue()
