#
# File: StagingTool.py
#
# Copyright (c) 2005, Pilot Systems (http://www.pilotsystems.net)
#
# GNU General Public Licence (GPL)
# 
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA  02111-1307  USA
#
"""Configuration item for StagingAddOn.

$Id: configuration.py 34913 2006-12-20 17:10:24Z dbaty $
"""

from AccessControl import ClassSecurityInfo

from Products.CMFCore.utils import getToolByName
from Products.CMFCore.permissions import View, ModifyPortalContent, ManagePortal

from Products.Archetypes.public import *
from Products.ATContentTypes.content.base import ATCTContent
from Products.ATContentTypes.content.schemata import finalizeATCTSchema

from Products.StagingAddOn.config import NEW_TRANSITION_ID
from Products.StagingAddOn.config import NEW_TRANSITION_PROPERTIES, PUBLISH_TRANSITION_POST_SCRIPT_ID, WORKFLOW_SCRIPTS

SECOND_SCHEMATA = '2'

## We want the user to select a workflow, and then select two
## transitions and a state of this workflow. This could be achieved in
## one form by using MasterSelectionWidget, but this would add a
## dependency only for the configuration of the product, which does
## not appear to me as something really necessary. (Damien)
schema = BaseSchema.copy() + Schema((
    StringField('id',
                required=True,
                vocabulary='getPossibleWorkflowIds',
                widget=SelectionWidget(format='select',
                                       label='Workflow',
                                       label_msgid='label_config_id',
                                       description="Select the workflow on which you want to "\
                                       "install StagingAddOn.",
                                       description_msgid='help_config_id',
                                       i18n_domain="stagingaddon",
                                       ),
                ),

    LinesField('publishedStateId',
               schemata=SECOND_SCHEMATA,
               required=True,
               vocabulary='getPossiblePublishedStateIds',
               widget=MultiSelectionWidget(format='select',
                                           label='Published state',
                                           label_msgid='label_config_published_state_id',
                                           description="Select the state from which you want " \
                                           "to be able to create new versions.",
                                           description_msgid='help_config_published_state_id',
                                           i18n_domain="stagingaddon",
                                           ),
               ),

    LinesField('publishTransitionId',
               schemata=SECOND_SCHEMATA,
               required=True,
               vocabulary='getPossiblePublishTransitionId',
               widget=MultiSelectionWidget(format='select',
                                           label='Publish transition',
                                           label_msgid='label_config_publish_transition_id',
                                           description="Select the transition which will " \
                                           "trigger the replacement of the previously " \
                                           "published version by the new version.",
                                           description_msgid='help_config_publish_transition_id',
                                           i18n_domain="stagingaddon",
                                           ),
               ),

    StringField('archiveTransitionId',
                schemata=SECOND_SCHEMATA,
                required=False,
                vocabulary='getPossibleArchiveTransitionId',
                widget=SelectionWidget(format='select',
                                       label='Archive transition',
                                       label_msgid='label_config_archive_transition_id',
                                       description="Select the transition which will " \
                                       "be triggered on the previously published item "\
                                       "when it is replaced by the new version.",
                                       description_msgid='help_config_archive_transition_id',
                                       i18n_domain="stagingaddon",
                                       ),
                ),

    BooleanField('expirePreviouslyPublishedVersion',
                 schemata=SECOND_SCHEMATA,
                 widget=BooleanWidget(label='Expire previously published version',
                                      label_msgid='label_config_expire',
                                      description="Toggle whether the previously published" \
                                      "version should be expired when the new version " \
                                      "replaces it.",
                                      description_msgid='help_config_expire',
                                      i18n_domain="stagingaddon",
                                      ),
                    ),
    ))


## Hide 'title' field
schema['title'].widget.visible= {'edit':'hidden', 'view': 'invisible'}

## Remove useless actions
useless = ('metadata', 'local_roles')
actions = list(ATCTContent.actions)
actions = [a for a in actions if a['id'] not in useless]
StagingAddOnConfigurationActions = tuple(actions)

class StagingAddOnConfiguration(ATCTContent):
    """Represent the staging configuration for a specific workflow."""
    __implements__ = ATCTContent.__implements__

    ## Type & icon
    meta_type = portal_type = 'StagingAddOnConfiguration'
    archetype_name = 'StagingAddOn configuration'
    content_icon = 'sao_config_icon.gif'

    ## Views
    default_view = 'configuration_view'
    immediate_view = default_view

    ## Various settings
    _at_rename_after_creation = False
    global_allow = False
    allow_discussion = False

    ## Actions
    actions = StagingAddOnConfigurationActions

    ## Security shortcuts
    security = ClassSecurityInfo()
    decPrivate = security.declarePrivate
    decProtected = security.declareProtected
    decPublic = security.declarePublic

    schema = schema

    #################################################################
    ## Various "magic" and utility methods, totally unrelated to
    ## staging features
    ###################
    def exclude_from_nav(self): return True

    ## Note that we do want "Configuration" items to be indexed. If
    ## not, they will not appear in the "Staging tool" folder listing
    ## and folder contents. And I do not want a specific view for it.

    decProtected(View, 'Title')
    def Title(self):
        """Return a semi-static i18n-ed title."""
        title = ''
        try:
            wf_title = self.getWorkflowTitleAndId()
            default = 'Configuration for %s'
            title = self.translate('configuration_title',
                                   domain='stagingaddon',
                                   default=default) % wf_title
        except AttributeError:
            pass ## No acquisition. Nevermind.
        return title


    decProtected(View, 'getWorkflowTitleAndId')
    def getWorkflowTitleAndId(self, wf_id=None):
        """Return the title and the id of ``wf_id`` as a single string.

        The returned string is built like this::

            <id> (<title>)
        """
        if wf_id is None:
            wf_id = self.getId()
        wtool = getToolByName(self, 'portal_workflow')
        wf = wtool.getWorkflowById(wf_id)
        if wf is None:
            return ''
        return '%s (%s)' % (wf.getId(), wf.title_or_id())
    #################################################################



    #################################################################
    ## Installation and uninstallation of the staging features
    ##########################################################
    decProtected(ModifyPortalContent, 'processForm')
    def processForm(self, *args, **kwargs):
        """Reinstall the configuration in case it has been modified."""
        ## Since we have two schematas, we have two forms. Therefore
        ## this method is called twice, including when the first form
        ## is submitted. At this moment, we should not do anything
        ## since we do not know the entire configuration.
        know_entire_configuration = self.getPublishedStateId() or \
            self.REQUEST.has_key('publishedStateId')
        if know_entire_configuration:
            self.uninstall()
        ATCTContent.processForm(self, *args, **kwargs)
        if know_entire_configuration:
            self.install()


    decPrivate('manage_beforeDelete')
    def manage_beforeDelete(self, item, container):
        """Uninstall configuration when item is removed."""
        self.uninstall()
        ATCTContent.manage_beforeDelete(self, item, container)


    decProtected('View', 'isInstalled')
    def isInstalled(self):
        """Return ``True`` iff the configuration has been installed."""
        return getattr(self.aq_base, '_installed', False)


    decProtected(ManagePortal, 'install')
    def install(self):
        """Install needed scripts and transitions in the workflow."""
        wtool = getToolByName(self, 'portal_workflow')

        ## Get workflow (and check that it exists)
        wf = wtool.getWorkflowById(self.getId())
        if wf is None:
            msg = "There is no '%s' workflow. Installation "\
            "has been cancelled." % self.getId()
            raise RuntimeError, msg

        self.installWorkflowScripts(wf)
        self.installTransition(wf)
        self.installScriptOnPublish(wf)
        self._installed = True


    decPrivate('installWorkflowScripts')
    def installWorkflowScripts(self, wf):
        """Install workflow scripts on workflow ``wf``."""
        for script_infos in WORKFLOW_SCRIPTS:
            if script_infos['id'] in wf.scripts.objectIds():
                msg = "Workflow '%s' already provides "\
                    "'%s' script. Installation has been "\
                    "cancelled." % (wf.getId(),
                                    script_infos['id'])
                raise RuntimeError, msg
            factory =  wf.scripts.manage_addProduct['PythonScripts']
            factory.manage_addPythonScript(id=script_infos['id'])
            script = wf.scripts[script_infos['id']]
            script.ZPythonScript_edit(params=script_infos['params'],
                                      body=script_infos['body'])


    decPrivate('installTransition')
    def installTransition(self, wf):
        """Install "*create new version*" transition and "attach" it to
        configured "*published*" states of workflow ``wf``.
        """
        if NEW_TRANSITION_ID in wf.transitions.objectIds():
            msg = "Workflow '%s' already provides "\
                "'%s' transition. Installation has been "\
                "cancelled." % (wf.getId(),
                                NEW_TRANSITION_ID)
            raise RuntimeError, msg
        wf.transitions.addTransition(NEW_TRANSITION_ID)
        transition = wf.transitions[NEW_TRANSITION_ID]
        transition.setProperties(**NEW_TRANSITION_PROPERTIES)

        for published_id in self.getPublishedStateId():
            published = getattr(wf.states.aq_base,
                                published_id,
                                None)
            if published is None:
                msg = "Workflow '%s' does not provide any "\
                "'%s' state. Installation has been "\
                "cancelled." % (wf.getId(),
                                published_id)
                raise RuntimeError, msg
            possible_transitions = list(published.transitions)
            if NEW_TRANSITION_ID not in possible_transitions:
                possible_transitions.append(NEW_TRANSITION_ID)
                published.transitions = tuple(possible_transitions)
        

    decPrivate('installScriptOnPublish')
    def installScriptOnPublish(self, wf):
        """Install script on configured "*publish*" transitions of workflow
        ``wf``.
        """
        for publish_id in self.getPublishTransitionId():
            transition = getattr(wf.transitions,
                                 publish_id,
                                 None)
            if transition is None:
                msg = "Workflow '%s' does not provide "\
                      "'%s' transition. Installation "\
                      "cancelled" % (self.getId(), publish_id)
                raise RuntimeError, msg
            if transition.script_name:
                msg = "Transition '%s' of workflow '%s' already provides "\
                      "a script ('%s'). Installation has been "\
                      "cancelled" % (transition.getId(),
                                     wf.getId(),
                                     transition.script_name)
                raise RuntimeError, msg
            transition.script_name = PUBLISH_TRANSITION_POST_SCRIPT_ID


    decProtected(ManagePortal, 'uninstall')
    def uninstall(self):
        """Uninstall scripts and transitions from the workflow."""
        wtool = getToolByName(self, 'portal_workflow')

        wf = wtool.getWorkflowById(self.getId())
        if wf is None:
            msg = "There is no '%s' workflow. Uninstallation "\
            "has been cancelled." % self.getId()
            raise RuntimeError, msg

        self.uninstallWorkflowScripts(wf)
        self.uninstallTransition(wf)
        self.uninstallScriptOnPublish(wf)
        self._installed = False


    decPrivate('uninstallWorkflowScripts')
    def uninstallWorkflowScripts(self, wf):
        """Uninstall workflow scripts on workflow ``wf``."""
        remove = []
        for script_infos in WORKFLOW_SCRIPTS:
            if script_infos['id'] in wf.scripts.objectIds():
                remove.append(script_infos['id'])
        wf.scripts.manage_delObjects(remove)


    decPrivate('uninstallTransition')
    def uninstallTransition(self, wf):
        """Uninstall "*create new version*" transition and "detach" it from
        configured "*published*" states.
        """
        if NEW_TRANSITION_ID in wf.transitions.objectIds():
            wf.transitions.manage_delObjects([NEW_TRANSITION_ID])

        for published_id in self.getPublishedStateId():
            published = getattr(wf.states.aq_base,
                                published_id,
                                None)
            if published is not None:
                possible_transitions = list(published.transitions)
                if NEW_TRANSITION_ID in possible_transitions:
                    possible_transitions.remove(NEW_TRANSITION_ID)
                    published.transitions = tuple(possible_transitions)


    decPrivate('uninstallScriptOnPublish')
    def uninstallScriptOnPublish(self, wf):
        """Uninstall script from configured "*publish*" transitions of
        workflow ``wf``.
        """
        for publish_id in self.getPublishTransitionId():
            transition = getattr(wf.transitions,
                                 publish_id,
                                 None)
            if transition is not None:
                if transition.script_name == PUBLISH_TRANSITION_POST_SCRIPT_ID:
                    transition.script_name = None
    ########################################################



    ########################################################
    ## Vocabulary methods
    #####################
    decPublic('getPossibleWorkflowIds')
    def getPossibleWorkflowIds(self):
        """Return ids of workflows on which StagingAddOn can be installed."""
        wtool = getToolByName(self, 'portal_workflow')
        available = list(wtool.getWorkflowIds())
        stool = self.getParentNode()
        for config in stool.objectValues(self.meta_type):
            if config.getId() is self.getId():
                continue
            if config.getId() in available:
                available.remove(config.getId())
        return [(wf_id, self.getWorkflowTitleAndId(wf_id)) \
                for wf_id in available]


    decPublic('getPossiblePublishedStateIds')
    def getPossiblePublishedStateIds(self):
        """Return id of the states of the selected workflow."""
        wtool = getToolByName(self, 'portal_workflow')
        wf = wtool[self.getId()]
        return wf.states.objectIds()


    decPublic('getPossiblePublishTransitionId')
    def getPossiblePublishTransitionId(self):
        """Return id of the transitions of the selected workflow."""
        wtool = getToolByName(self, 'portal_workflow')
        wf = wtool[self.getId()]
        transitions = wf.transitions.objectIds()
        if NEW_TRANSITION_ID in transitions:
            transitions.remove(NEW_TRANSITION_ID)
        return transitions


    decPublic('getPossibleArchiveTransitionId')
    def getPossibleArchiveTransitionId(self):
        """Return id of the transitions of the selected workflow."""
        wtool = getToolByName(self, 'portal_workflow')
        wf = wtool[self.getId()]
        transitions = wf.transitions.objectIds()
        if NEW_TRANSITION_ID in transitions:
            transitions.remove(NEW_TRANSITION_ID)
        transitions = [(t, t) for t in transitions]
        none = self.translate('no_transition',
                              domain='stagingaddon',
                              default='None')
        transitions.insert(0, ('', none))
        return transitions
    ########################################################

registerType(StagingAddOnConfiguration)
