# -*- coding: utf-8 -*-
# File  : StagingTool.py
#
# Copyright (c) 2005, Pilot Systems (http://www.pilotsystems.net)
#
# GNU General Public Licence (GPL)
# 
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place, Suite 330, Boston, MA  02111-1307  USA
#
"""Define the main tool used by StagingAddOn.

$Id: staging_tool.py 42837 2007-05-29 20:45:10Z dbaty $
"""

import logging
from types import ListType

import transaction
from DateTime import DateTime
from zExceptions import Redirect
from Globals import InitializeClass
from persistent.list import PersistentList
from OFS.PropertyManager import PropertyManager
from AccessControl import ClassSecurityInfo, getSecurityManager, \
    Unauthorized
from AccessControl.SecurityManagement import newSecurityManager, \
    setSecurityManager
from AccessControl.User import UnrestrictedUser as BaseUnrestrictedUser

from Products.CMFCore.utils import getToolByName
from Products.CMFCore.permissions import ManagePortal
from Products.CMFCore.WorkflowCore import WorkflowException

from Products.Archetypes.public import registerType
from Products.Archetypes.config import REFERENCE_CATALOG, UID_CATALOG
from Products.ATContentTypes.content.folder import ATFolder, ATFolderSchema

from Products.StagingAddOn.config import TESTING
from Products.StagingAddOn.exceptions import NotCurrentVersion, NewVersionAlreadyExists, NotConfiguredForSAO, NotPublished


ID = 'portal_staging'
TITLE = 'Staging tool'
META_TYPE = 'StagingAddOnTool'

LOG = logging.getLogger('StagingAddOn')

class UnrestrictedUser(BaseUnrestrictedUser):
    """Unrestricted user that still has an id."""
    def getId(self):
        """Return the ID of the user."""
        return self.getUserName()


## Remove useless actions
useless = ('metadata', 'local_roles')
actions = list(ATFolder.actions)
actions = [a for a in actions if a['id'] not in useless]
StagingToolActions = tuple(actions)

class StagingTool(ATFolder):
    """This is the main tool used by StagingAddOn."""
    __implements__ = ATFolder.__implements__

    ## Type & icon
    meta_type = portal_type = META_TYPE
    archetype_name = 'Staging tool'
    content_icon = 'staging_tool.gif'

    ## Views
    default_view = 'base_view'
    immediate_view = default_view
    suppl_views = ()

    ## Various settings
    _at_rename_after_creation = False
    global_allow = False
    allow_discussion = False
    filter_content_types = True
    allowed_content_types = ('StagingAddOnConfiguration', )

    ## Actions
    actions = StagingToolActions

    schema = ATFolderSchema.copy()

    security = ClassSecurityInfo()
    decPrivate = security.declarePrivate
    decProtected = security.declareProtected
    decPublic = security.declarePublic


    #################################################################
    ## Various "magic", totally unrelated to staging features
    #########################################################
    def canSetDefaultPage(self): return False

    def exclude_from_nav(self): return True

    def _setId(self, id):
        if id != self.getId():
            raise RuntimeError, 'No way!'

    ## We do not want the tool to be indexed.
    def doNothing(self, *args, **kwargs): pass
    indexObject = unindexObject = reindexObject = doNothing
    index_object = unindex_object = reindex_object = doNothing

    def Title(self):
        title = ''
        try:
            title = self.translate('staging_tool_title',
                                   domain='plone',
                                   default='Configuration of StagingAddOn')
        except AttributeError:
            pass ## No acquisition. Nevermind.
        return title
    #################################################################


    #################################################################
    ## Create and promote versions
    ##############################
    decPublic('createNewVersion')
    def createNewVersion(self, obj, redirect=True):
        """Create a new version for ``obj`` and return the created object.

        If ``redirect`` is ``True``, the method redirects to the new
        version. (In this case, it does not return anything.)

        If the current user cannot create an object in the same
        container as ``obj``, the ``Unauthorized`` exception will be
        raised.

        If ``obj`` is not the current version, the
        ``NotCurrentVersion`` exception is raised.

        If a new version has already been created, the
        ``NewVersionAlreadyExists`` exception is raised.
        """
        self.isAllowedToCreateNewVersion(obj, raise_exceptions=True)
        container = obj.aq_inner.aq_parent
        new_version_id = self._generateIdForNewVersion(obj.getId())
        new_version = self._duplicateObject(obj, new_version_id)
        new_version._version = self.getVersionNumber(obj) + 1
        if getattr(new_version.aq_base, '_new_version', False):
            del new_version._new_version
        if getattr(new_version.aq_base, '_versions', False):
            del new_version._versions
        new_version._is_a_version_of = self._getUID(obj)
        new_version_uid = self._getUID(new_version)
        if not getattr(obj.aq_base, '_versions', False):
            obj._versions = PersistentList()
        obj._versions.append(new_version_uid)
        obj._new_version = new_version_uid

        ## Put the new version just after the current one.
        position = 1 + container.getObjectPosition(obj.getId())
        container.moveObjectToPosition(new_version_id,
                                       position)

        if redirect and not TESTING:
            ## We have to raise a Redirect exception so that this
            ## method can be called from a workflow script. The
            ## "normal" 'response.redirect()' way does not work in
            ## this case. But if we do not manually commit the
            ## transaction, Zope will not commit it by itself because
            ## of the exception we raise.
            ## Obviously, we do not want to commit anything when
            ## running tests.
            transaction.commit()
            msg = 'currently_editing_new_version'
            qs = 'portal_status_message=%s' % msg
            url = '%s/edit?%s' % (new_version.absolute_url(), qs)
            raise Redirect, url
        return new_version


    decPublic('isAllowedToCreateNewVersion')
    def isAllowedToCreateNewVersion(self, obj,
                                    raise_exceptions=False):
        """Return ``True`` iff the current user is allowed to create a new
        version of ``obj``.

        If ``raise_exceptions`` is ``True``, exceptions may be raisen.
        """
        config = self.getConfigFor(obj)
        if config is None:
            raise NotConfiguredForSAO

        ## Is 'obj' in one of the configured 'published' workflow
        ## states?
        wtool = getToolByName(self, 'portal_workflow')
        state = wtool.getInfoFor(obj, 'review_state')
        if state not in config.getPublishedStateId():
            if raise_exceptions:
                raise NotPublished
            return False

        ## Is 'obj' the current version?
        if not self.isCurrentVersion(obj):
            if raise_exceptions:
                raise NotCurrentVersion
            return False

        ## Does 'obj' already have a new version?
        if self.hasNewVersion(obj):
            if raise_exceptions:
                raise NewVersionAlreadyExists
            return False

        ## Is 'obj' a version of another item?
        if getattr(obj, '_is_a_version_of', False):
            if raise_exceptions:
                raise NotCurrentVersion
            return False

        ## Can the current user add a new item of the same content
        ## type?
        container = obj.aq_inner.aq_parent
        ttool = getToolByName(obj, 'portal_types')
        fti = ttool.getTypeInfo(obj.portal_type)
        if not fti.isConstructionAllowed(container):
            if raise_exceptions:
                raise Unauthorized
            return False
        return True


    decPublic('promoteVersion')
    def promoteVersion(self, obj, comment=''):
        """Swap currently published and new (``obj``) versions by renaming
        both, archive the currently published one, and return a tuple
        which contains the new published version and the old one.

        If the current version cannot be found, we do nothing and
        return ``None``.
        """
        config = self.getConfigFor(obj)
        if config is None:
            raise NotConfiguredForSAO

        uid_handler = getToolByName(self, 'portal_uidhandler')
        brain = uid_handler.queryBrain(obj._is_a_version_of,
                                       extra_criteria={'Language': 'all',
                                                       'show_inactive': 1})
        if brain is None:
            return obj
        currently_published = brain.getObject()
        currently_published_id = currently_published.getId()

        ## The current user may not be allowed to do one of the
        ## actions below. We hope that a 'Manager' is, so we
        ## temporarily change the security context to use a temporary
        ## 'Manager' user.
        portal = getToolByName(self, 'portal_url').getPortalObject()
        old_sm = getSecurityManager()
        tmp_user = UnrestrictedUser(old_sm.getUser().getId(),
                                    '', ['Manager'], '')
        tmp_user = tmp_user.__of__(portal.acl_users)
        newSecurityManager(None, tmp_user)

        ## First, rename the currently published version
        old_id = self._generateIdForOldVersion(currently_published_id)
        container = obj.aq_inner.aq_parent
        container.manage_renameObjects((currently_published_id, ),
                                       (old_id, ))

        ## Archive it if the configuration says so
        old = container[old_id]
        action = config.getArchiveTransitionId()
        wtool = getToolByName(self, 'portal_workflow')
        if action:
            try:
                wtool.doActionFor(old, action, comment=comment)
            except WorkflowException:
                pass ## Silently fail if 'action' is not available.

        ## Expire it if the configuration says so
        if config.expirePreviouslyPublishedVersion:
            now = DateTime()
            old.setExpirationDate(now)
            if TESTING:
                ## 'expirationDate' is a DateRangeIndex, and this type
                ## of index does not store very precise time
                ## informations (for performance issues, mainly). It
                ## results that, in an automated test, an item which
                ## has been very recently expired (less than one
                ## second before, for example), is not seen as expired
                ## by the catalog. Thus, we do not have the same
                ## behaviour in the tests as in a real (human)
                ## usage. Since testing should mimic as well as
                ## possible real usage, we have to take care of this
                ## problem. (If we did not, we could break the code
                ## and still have tests not failing.)
                ## That is why we here expire the item "more" than we
                ## usually do. Thus, the catalog does see this item as
                ## expired. (Damien)
                old.setExpirationDate(now - 1)
            old.reindexObject(('expirationDate', ))

        ## Rename 'obj'
        container.manage_renameObjects((obj.getId(), ),
                                       (currently_published_id, ))

        ## Reset security manager
        setSecurityManager(old_sm)

        ## Swap old and new UIDs so that references to the previous
        ## version are changed to references to the newly published
        ## one.
        old_uid = old.UID()
        new_uid = obj.UID()
        old._setUID(new_uid)
        obj._setUID(old_uid)

        ## Update specific attributes
        del obj._is_a_version_of
        del old._new_version
        obj._versions = old._versions[:]
        del old._versions
        obj._versions[-1] = self._getUID(old)
        obj_uid = self._getUID(obj)
        old._is_a_version_of = obj_uid
        for uid in obj._versions:
            brain = uid_handler.unrestrictedQueryBrain(uid)
            if brain is not None:
                o = brain._unrestrictedGetObject()
                o._is_a_version_of = obj_uid

        return (container[currently_published_id],
                container[old_id])
    #################################################################


    #################################################################
    ## Utility methods
    ##################
    decPublic('showPortlet')
    def showPortlet(self, obj):
        """Return ``True`` iff the staging portlet should be shown on ``obj``.

        It checks if there is a staging configuration for the workflow
        of the displayed content.
        """
        portal = getToolByName(self, 'portal_url').getPortalObject()
        if obj == portal:
            ptool = getToolByName(self, 'plone_utils')
            default_page_id = ptool.getDefaultPage(obj)
            if default_page_id is None:
                return False
            obj = portal[default_page_id]
        return self.getConfigFor(obj) is not None


    decPublic('getConfigFor')
    def getConfigFor(self, obj):
        """Return the StagingAddOn configuration for ``obj`` or ``None`` if it
        cannot be found.
        """
        wtool = getToolByName(self, 'portal_workflow')
        workflows = wtool.getWorkflowsFor(obj)
        if not workflows:
            return None
        ## We only take care of the first workflow.
        workflow = workflows[0].getId()
        return getattr(self.aq_explicit, workflow, None)


    decPublic('getVersions')
    def getVersions(self, obj, sortFunc=None):
        """Return versions of ``obj`` which are accessible by the current
        user.

        All versions of ``obj`` are returned, including ``obj``
        itself. If ``sortFunc`` is provided, the method uses it to
        sort the items.

        **Warning:** only viewable items are returned.
        """
        if not self.isCurrentVersion(obj):
            current_version = self._getCurrentVersion(obj, True)
            if current_version is None:
                return ()
            return self.getVersions(current_version, sortFunc)

        ## Ok, this is the current version
        uid_handler = getToolByName(self, 'portal_uidhandler')
        versions = []
        for uid in getattr(obj, '_versions', ()):
            extra_criteria = {'show_inactive': True}
            brain = uid_handler.queryBrain(uid, extra_criteria=extra_criteria)
            if brain is None:
                continue
            versions.append(brain.getObject())

        mtool = getToolByName(self, 'portal_membership')
        if mtool.checkPermission('View', obj):
            ## Only include 'obj' if it is viewable by the current
            ## user.
            versions.append(obj)
        if sortFunc is not None:
            versions.sort(sortFunc)
        return versions


    decPublic('getStatus')
    def getStatus(self, obj):
        """Return status of ``obj``.

        Status can be "*current version*", "*new version*" or "*old
        version*".
        """
        if self.isAnOldVersion(obj):
            return 'old version'
        elif self.isTheNewVersion(obj):
            return 'new version'
        return 'current version'


    decPublic('isCurrentVersion')
    def isCurrentVersion(self, obj):
        """Return ``True iff ``obj`` is the the current version."""
        return not getattr(obj.aq_base, '_is_a_version_of', False)


    decPublic('isTheNewVersion')
    def isTheNewVersion(self, obj):
        """Return ``True`` iff ``obj`` is a new version of another item."""
        current_version = self._getCurrentVersion(obj, True)
        if current_version is None:
            return False
        new_version_uid = getattr(current_version.aq_base,
                                  '_new_version',
                                  None)
        return new_version_uid == self._getUID(obj)


    decPublic('isAnOldVersion')
    def isAnOldVersion(self, obj):
        """Return ``True`` iff ``obj`` is an old version of another item."""
        current_version = self._getCurrentVersion(obj, True)
        if current_version is None:
            return False
        if obj == current_version:
            return False
        new_version_uid = getattr(current_version.aq_base,
                                  '_new_version',
                                  None)
        return self._getUID(obj) != new_version_uid


    decPublic('hasNewVersion')
    def hasNewVersion(self, obj):
        """Return ``True`` iff ``obj`` has a new version.

        This method actually checks if the reference to the new
        version still references an object.
        """
        return self._getNewVersion(obj, unrestricted=True) is not None


    decPublic('getNewVersion')
    def getNewVersion(self, obj):
        """Return the new version of ``obj`` or ``None`` if none exists or if
        the current user cannot access it.
        """
        return self._getNewVersion(obj)


    decPrivate('_getNewVersion')
    def _getNewVersion(self, obj, unrestricted=False):
        """Return the new version of ``obj' or ``None`` if none exists or if
        the current user cannot access it.

        If ``unrestricted`` is ``True``, no security check is done: if
        a new version exists, it is returned, even if the current user
        cannot access it.
        """
        uid = getattr(obj.aq_base, '_new_version', None)
        if uid is None:
            return None

        uid_handler = getToolByName(self, 'portal_uidhandler')
        if unrestricted:
            query = uid_handler.unrestrictedQueryBrain
        else:
            query = uid_handler.queryBrain

        new_version_brain = query(uid)
        if new_version_brain is None:
            if unrestricted:
                ## The referenced item is definitely deceased (and no,
                ## the plumage does not enter into it). We remove this
                ## reference.
                obj._new_version = None
            return None
        return new_version_brain._unrestrictedGetObject()


    decPublic('getCurrentVersion')
    def getCurrentVersion(self, obj):
        """Return the current version of ``obj``, or ``None`` if none could be
        found.
        """
        return self._getCurrentVersion(obj)


    decPrivate('_getCurrentVersion')
    def _getCurrentVersion(self, obj, unrestricted=False):
        """Return the current version of ``obj``, or ``None`` if the
        referenced one could not be found or if the current user
        cannot access it.

        If ``unrestricted`` is ``True``, no security check is done: if
        the current version exists, it is returned, even if the
        current user cannot access it.
        """
        uid = getattr(obj.aq_base, '_is_a_version_of', None)
        if uid is None:
            mtool = getToolByName(self, 'portal_membership')
            if mtool.checkPermission('View', obj):
                return obj
            return None

        uid_handler = getToolByName(self, 'portal_uidhandler')
        if unrestricted:
            query = uid_handler.unrestrictedQueryBrain
        else:
            query = uid_handler.queryBrain

        current_version_brain = query(uid)
        if current_version_brain is None:
            if unrestricted:
                ## The referenced item is no more (and no, it would
                ## not vooooom). We remove this reference.
                obj._is_a_version_of = None
            return None
        return current_version_brain._unrestrictedGetObject()


    decPublic('getVersionNumber')
    def getVersionNumber(self, obj):
        """Return the version number of ``obj``.

        Default is 1.
        """
        return getattr(obj.aq_base, '_version', 1)
    #################################################################


    #################################################################
    ## Installation shortcuts
    #########################
    decProtected(ManagePortal, 'installStagingFor')
    def installStagingFor(self, workflow_id='plone_workflow',
                          published_id=('published', ),
                          publish_id=('publish', ),
                          archive_id='',
                          expires=False):
        """Install staging configuration for ``workflow_id``."""
        oid = self.invokeFactory('StagingAddOnConfiguration',
                                 workflow_id)
        config = self[oid]
        config.processForm(data=1, values={'id': workflow_id})
        request = {'publishedStateId': list(published_id),
                   'publishTransitionId': list(publish_id),
                   'archiveTransitionId': archive_id,
                   'expirePreviouslyPublishedVersion': expires}
        config.REQUEST = request
        config.processForm(data=1, values=request)


    decProtected(ManagePortal, 'uninstallStagingFor')
    def uninstallStagingFor(self, workflow_id='plone_workflow'):
        """Remove staging configuration for ``workflow_id``."""
        self.manage_delObjects([workflow_id])
    #################################################################


    #################################################################
    ## Internal methods
    ###################
    def _duplicateObject(self, obj, new_id):
        """Duplicate object, try to rename it with ``new_id`` and return the
        copy.

        The current way of duplicating the object is:

        1. Copy and paste the object.

        2. Try to rename the copy as requested.

        3. Manually copy value of reference fields.

        4. Set creation date of the copy.

        5. Remove workflow history of the copy.
        """
        container = obj.aq_inner.aq_parent
        cp = container.manage_copyObjects(obj.getId())
        result = container.manage_pasteObjects(cp)

        generated_new_id = result[0]['new_id']
        container.manage_renameObjects((generated_new_id, ),
                                       (new_id, ))

        copy = container[new_id]

        ## Values of reference fields are not kept in Archetypes by
        ## default. This is a major problem for StagingAddOn, so we
        ## manually copy values (and we recursively do so if the item
        ## is folderish.
        if obj.is_folderish():
            def walk(tool, orig_folder, copy_folder):
                tool._copyReferenceFieldValues(orig_folder,
                                               copy_folder)
                for orig_item in orig_folder.objectValues():
                    copy_item = copy_folder[orig_item.getId()]
                    if orig_item.is_folderish():
                        walk(tool, orig_item, copy_item)
                    else:
                        tool._copyReferenceFieldValues(orig_item,
                                                       copy_item)
            walk(self, obj, copy)
        else:
            self._copyReferenceFieldValues(obj, copy)

        self._removeWorkflowHistory(copy)
        copy.setCreationDate(DateTime())
        copy.reindexObject(('created', ))
        return copy


    def _copyReferenceFieldValues(self, orig, copy):
        """Copy values of ``ReferenceField``s of ``orig`` to ``copy``.

        ``orig`` and ``copy`` should obvisouly have the same schema.
        """
        indexes = []
        for field in orig.Schema().filterFields(type='reference'):
            value = field.get(orig)
            getattr(copy, field.mutator)(value)
            indexes.append(field.getName())
        if indexes:
            copy.reindexObject(indexes)


    def _getUID(self, obj):
        """Return an UID for ``obj``.

        This is only a wrapper around CMFUid's ``portal_uidhandler``.
        """
        portal_uidhandler = getToolByName(self, 'portal_uidhandler')
        return portal_uidhandler.register(obj)


    def _generateIdForNewVersion(self, current_id):
        """Generate an id for the new version of an object whose id is
        ``current_id``.
        """
        gen = self.translate('%s-new-version', domain='stagingaddon')
        gen = gen % current_id
        ## If we do not do that, we pass an Unicode object, and
        ## 'ObjectManager._checkValidId()' does not allow it.
        gen = str(gen)
        return gen


    def _generateIdForOldVersion(self, current_id):
        """Generate an id for an old version of an object whose id is
        ``current_id``.
        """
        gen = self.translate('%s-old-version-replaced-%s',
                             domain='stagingaddon')
        gen = gen % (current_id, DateTime().strftime('%d%m%Y-%H%M%S'))
        ## If we do not do that, we pass an Unicode object, and
        ## 'ObjectManager._checkValidId()' does not allow it.
        gen = str(gen)
        return gen


    def _removeWorkflowHistory(self, obj):
        """Remove workflow history of ``obj``."""
        wtool = getToolByName(self, 'portal_workflow')
        workflows = wtool.getWorkflowsFor(obj)
        if not workflows:
            return
        ## We only take care of the first workflow.
        wf_id = workflows[0].getId()
        history = list(obj.workflow_history[wf_id])
        obj.workflow_history[wf_id] = (history[-1], )
    #################################################################
    
registerType(StagingTool)
