"""
Tests for StagingAddOn configuration

$Id: testConfiguration.py 27075 2006-08-11 08:18:40Z dbaty $
"""

import os, sys
if __name__ == '__main__':
    execfile(os.path.join(sys.path[0], 'framework.py'))

from Products.StagingAddOn.tests.utils import StagingAddOnTestCase

import transaction
from zExceptions import BadRequest
from AccessControl import Unauthorized
from Products.DCWorkflow.Guard import Guard
from Products.CMFCore.utils import getToolByName
from Products.StagingAddOn.content.staging_tool import ID as STOOL_ID
from Products.StagingAddOn.config import WORKFLOW_SCRIPTS, PUBLISH_TRANSITION_POST_SCRIPT_ID, NEW_TRANSITION_ID, NEW_TRANSITION_PROPERTIES

class TestConfiguration(StagingAddOnTestCase):
    """
    Make sure that the configuration really configures workflows as it
    should

    These tests do not test staging features.
    """

    def afterSetUp(self):
        """
        Called before each tests

        - create users and content;
        - install a default staging configuration for 'plone_workflow'
        """
        portal = self.portal

        self.createTestUsersAndContent()

        self.login('manager')
        stool = getToolByName(portal, STOOL_ID)
        stool.installStagingFor()

        ## Finally, log out.
        self.logout()


    def testInstallation(self):
        """
        Test if the installation of a configuration does install
        things correctly
        """
        portal = self.portal
        wtool = getToolByName(portal, 'portal_workflow')
        stool = getToolByName(portal, STOOL_ID)
        document = portal.folder.document1

        self.login('manager')
        workflow_id = 'plone_workflow'
        published_id = ('published', )
        publish_id = ('publish', )
        wf = wtool.getWorkflowById(workflow_id)

        self.failUnless(stool.getConfigFor(document) == stool[workflow_id])

        ## Test that workflow scripts have been correctly installed.
        for script_infos in WORKFLOW_SCRIPTS:
            script = getattr(wf.scripts.aq_base,
                             script_infos['id'], None)
            self.failUnless(script is not None,
                            "Script '%s' could not be " \
                            "found" % script_infos['id'])
            ## An extra newline is added to the body. See
            ## 'PythonScript.__setstate__()'.
            self.failUnless(script.body() == script_infos['body'] + '\n')
            self.failUnless(script.params() == script_infos['params'])

        ## Test that the new transition has been correctly installed.
        transition = getattr(wf.transitions.aq_base,
                             NEW_TRANSITION_ID, None)
        self.failUnless(transition is not None)
        for key, value in NEW_TRANSITION_PROPERTIES.items():
            if key == 'props':
                g = Guard()
                g.changeFromProperties(value)
                self.failUnless(g.getSummary() == transition.guard.getSummary())
            else:
                v = getattr(transition, key, None)
                self.failUnless(v == value)

        ## Test that 'publish' transition has been correctly
        ## configured
        script_name = PUBLISH_TRANSITION_POST_SCRIPT_ID
        for id in publish_id:
            transition = wf.transitions[id]
            self.failUnless(transition.script_name == script_name)


    def testUninstallation(self):
        """
        Test that everything is removed when a staging configuration
        is uninstalled
        """
        portal = self.portal
        wtool = getToolByName(portal, 'portal_workflow')
        stool = getToolByName(portal, STOOL_ID)
        document = portal.folder.document1
        workflow_id = 'plone_workflow'
        published_id = ('published', )
        publish_id = ('publish', )
        wf = wtool.getWorkflowById(workflow_id)

        self.login('manager')
        stool.uninstallStagingFor()
        self.failUnless(stool.getConfigFor(document) is None)

        ## Test that workflow scripts have been removed
        for script_infos in WORKFLOW_SCRIPTS:
            self.failUnless(not getattr(wf.scripts.aq_base,
                                        script_infos['id'],
                                        False))

        ## Test that the new transition has been removed
        self.failUnless(NEW_TRANSITION_ID not in \
                        wf.transitions.objectIds())
        for id in published_id:
            published = wf.states[id]
            self.failUnless(NEW_TRANSITION_ID not in published.transitions)

        ## Test that 'publish' transition has been correctly
        ## reconfigured
        for id in publish_id:
            transition = wf.transitions[id]
            self.failUnless(transition.script_name == None)


    def testConfigurationRemoval(self):
        """
        Test that everything is removed when a staging configuration
        is removed via the PMI
        """
        portal = self.portal
        wtool = getToolByName(portal, 'portal_workflow')
        stool = getToolByName(portal, STOOL_ID)
        document = portal.folder.document1
        workflow_id = 'plone_workflow'
        published_id = ('published', )
        publish_id = ('publish', )
        wf = wtool.getWorkflowById(workflow_id)

        self.login('manager')
        stool.manage_delObjects([workflow_id])
        self.failUnless(stool.getConfigFor(document) is None)

        ## Test that workflow scripts have been removed
        for script_infos in WORKFLOW_SCRIPTS:
            self.failUnless(not getattr(wf.scripts.aq_base,
                                        script_infos['id'],
                                        False))

        ## Test that the new transition has been removed
        self.failUnless(NEW_TRANSITION_ID not in \
                        wf.transitions.objectIds())
        for id in published_id:
            published = wf.states[id]
            self.failUnless(NEW_TRANSITION_ID not in published.transitions)

        ## Test that 'publish' transition has been correctly
        ## reconfigured
        for id in publish_id:
            transition = wf.transitions[id]
            self.failUnless(transition.script_name == None)


    def testInstallationIsUnique(self):
        """
        Test that when a configuration has been installed for a
        workflow, it is not possible to install another configuration
        for the same workflow
        """
        portal = self.portal
        stool = getToolByName(portal, STOOL_ID)

        ## Trying to install again on the same workflow should raise
        ## 'BadRequest' exception.
        self.login('manager')
        self.assertRaises(BadRequest, stool.installStagingFor)


    def testModification(self):
        """
        Test that modifying a configuration really changes it
        """
        portal = self.portal
        wtool = getToolByName(portal, 'portal_workflow')
        stool = getToolByName(portal, STOOL_ID)
        folder = portal.folder
        document = folder.document1

        self.login('manager')

        config = stool.getConfigFor(document)
        old_workflow_id = config.getId()
        old_published_id = config.getPublishedStateId()
        old_publish_id = config.getPublishTransitionId()
        old_archive_id = config.getArchiveTransitionId()

        new_workflow_id = 'folder_workflow'
        new_published_id = ('visible', )
        new_publish_id = ('show', )
        new_archive_id = ''
        expires = True

        self.login('manager')

        ## Make sure that we have '_p_jar' on 'config', otherwise Zope
        ## will not let us rename it.
        transaction.savepoint()

        request = {'id': new_workflow_id}
        config.REQUEST = request
        config.processForm(data=1,
                           values=request)
        request = {'publishedStateId': list(new_published_id),
                   'publishTransitionId': list(new_publish_id),
                   'archiveTransitionId': new_archive_id,
                   'expirePreviouslyPublishedVersion': expires}
        config.REQUEST = request
        config.processForm(data=1, values=request)
        config.REQUEST = None

        ## There should not be any configuration for document, which
        ## uses 'plone_workflow'
        config = stool.getConfigFor(document)
        self.failUnless(config is None)

        ## There should be a configuration for a folder (which uses
        ## 'folder_workflow'), though.
        config = stool.getConfigFor(folder)
        self.failUnless(config is not None)

        ## Test that the new transition has been removed from
        ## 'old_published_id' state...
        wf = wtool[old_workflow_id]
        for id in old_published_id:
            published = wf.states[id]
            self.failUnless(NEW_TRANSITION_ID not in published.transitions)

        ## ... and that it has been added to the 'new_published_id'
        ## state.
        wf = wtool[new_workflow_id]
        for id in new_published_id:
            published = wf.states[id]
            self.failUnless(NEW_TRANSITION_ID in published.transitions)

        ## Test that 'old_publish_id' transition has been correctly
        ## reconfigured...
        wf = wtool[old_workflow_id]
        for id in old_publish_id:
            transition = wf.transitions[id]
            self.failUnless(transition.script_name == None)

        ## ... and that 'new_publish_id' transition has been correctly
        ## reconfigured, too.
        wf = wtool[new_workflow_id]
        for id in new_publish_id:
            transition = wf.transitions[id]
            self.failUnless(transition.script_name == PUBLISH_TRANSITION_POST_SCRIPT_ID)

        ## Test that workflow scripts have been removed from the old
        ## workflow...
        wf = wtool[old_workflow_id]
        for script_infos in WORKFLOW_SCRIPTS:
            self.failUnless(not getattr(wf.scripts.aq_base,
                                        script_infos['id'],
                                        False))

        ## ... and added to the new one
        wf = wtool[new_workflow_id]
        for script_infos in WORKFLOW_SCRIPTS:
            self.failUnless(getattr(wf.scripts.aq_base,
                                    script_infos['id'],
                                    False))

        ## Test that the new transition has been removed from the old
        ## workflow...
        wf = wtool[old_workflow_id]
        wf_transitions = wf.transitions.objectIds()
        self.failUnless(NEW_TRANSITION_ID not in wf_transitions)
        for id in old_published_id:
            published = wf.states[id]
            self.failUnless(NEW_TRANSITION_ID not in published.transitions)

        ## ... and added to the new one
        wf = wtool[new_workflow_id]
        wf_transitions = wf.transitions.objectIds()
        self.failUnless(NEW_TRANSITION_ID in wf_transitions)
        for id in new_published_id:
            published = wf.states[id]
            self.failUnless(NEW_TRANSITION_ID in published.transitions)


if  __name__ == '__main__':
    framework()

def test_suite():
    from unittest import TestSuite, makeSuite
    suite = TestSuite()
    suite.addTest(makeSuite(TestConfiguration))
    return suite
