"""
Tests for StagingAddOn StagingTool features

$Id: testStaging.py 42837 2007-05-29 20:45:10Z dbaty $
"""

import os, sys
if __name__ == '__main__':
    execfile(os.path.join(sys.path[0], 'framework.py'))

from Products.StagingAddOn.tests.utils import HAS_LINGUA_PLONE
from Products.StagingAddOn.tests.utils import StagingAddOnTestCase

from AccessControl import Unauthorized
from Products.CMFCore.utils import getToolByName
from Products.CMFCore.WorkflowCore import WorkflowException
from Products.StagingAddOn.config import NEW_TRANSITION_ID
from Products.StagingAddOn.content.staging_tool import ID as STOOL_ID
from Products.StagingAddOn.exceptions import NotCurrentVersion, NewVersionAlreadyExists, NotConfiguredForSAO, NotPublished


class TestStaging(StagingAddOnTestCase):
    """Make sure that the staging features work as expected"""

    def afterSetUp(self):
        """
        Called before each tests

        - create users and content;
        """
        portal = self.portal

        self.createTestUsersAndContent()

        self.login('manager')
        stool = getToolByName(portal, STOOL_ID)
        stool.installStagingFor()

        ## Finally, log out.
        self.logout()


    def testCreateVersion(self):
        """
        Test that creating a new version of an item does not crash
        your portal
        """
        portal = self.portal
        wtool = getToolByName(portal, 'portal_workflow')
        stool = getToolByName(portal, STOOL_ID)

        ## Not in one of the 'published' states
        self.login('manager')
        document = portal.folder.document1
        self.assertRaises(WorkflowException, wtool.doActionFor,
                          ob=document, action=NEW_TRANSITION_ID)
        self.assertRaises(NotPublished, stool.createNewVersion, document)

        ## Cannot add a new item in the same folder
        wtool.doActionFor(document, 'publish')
        self.login('member1')
        self.assertRaises(WorkflowException, wtool.doActionFor,
                          ob=document, action=NEW_TRANSITION_ID)
        self.assertRaises(Unauthorized, stool.createNewVersion, document)

        ## Create a new version
        document = portal.Members.member1.document
        self.login('manager')
        wtool.doActionFor(document, 'publish')
        self.login('member1')
        wtool.doActionFor(document, NEW_TRANSITION_ID)
        new_version_id = stool._generateIdForNewVersion(document.getId())
        new_version = document.aq_parent[new_version_id]
        self.failUnless(stool.getNewVersion(document) == new_version)

        ## Cannot create another new version
        self.assertRaises(WorkflowException, wtool.doActionFor,
                          ob=document, action=NEW_TRANSITION_ID)
        self.assertRaises(NewVersionAlreadyExists,
                          stool.createNewVersion, document)



    def testCanCreateANewVersionAfterHavingRemovedThePreviousOne(self):
        """
        Create a new version, remove it and then test that we can
        create a new version again
        """
        portal = self.portal
        wtool = getToolByName(portal, 'portal_workflow')
        stool = getToolByName(portal, STOOL_ID)
        folder = portal.folder
        document = folder.document1
        self.login('manager')
        wtool.doActionFor(document, 'publish')

        new_version = stool.createNewVersion(document, redirect=False)

        ## We cannot create a new version. It already exists.
        self.assertRaises(NewVersionAlreadyExists,
                          stool.createNewVersion, document)

        ## Remove the new version and try to create it again
        folder.manage_delObjects([new_version.getId()])
        new_version = stool.createNewVersion(document, redirect=False)


    def testCreateNewVersionKeepReferences(self):
        """
        Test that 'createNewVersion()' keeps references
        """
        portal = self.portal
        wtool = getToolByName(portal, 'portal_workflow')
        stool = getToolByName(portal, STOOL_ID)
        folder = portal.folder
        document = folder.document1
        document.setRelatedItems([folder.UID()])
        self.login('manager')
        wtool.doActionFor(document, 'publish')

        new_version = stool.createNewVersion(document, redirect=False)
        orig_value = document.getRelatedItems()
        new_value = new_version.getRelatedItems()
        self.failUnless(orig_value == new_value)


    def testReferencesAreConsistent(self):
        """Test that references to and from a newly published item are
        equivalent to those of the old version.
        """
        portal = self.portal
        wtool = getToolByName(portal, 'portal_workflow')
        stool = getToolByName(portal, STOOL_ID)
        sortFunc = lambda x, y: cmp(x.creation_date,
                                    y.creation_date)
        document = portal.folder.document1
        other_doc = portal.folder.subfolder.document2

        self.login('manager')

        ## Set a reference to ``document``
        other_doc.setRelatedItems([document.UID()])

        ## Create a new version and publish it
        wtool.doActionFor(document, 'publish')
        new_version = stool.createNewVersion(document, redirect=False)
        wtool.doActionFor(new_version, 'publish')
        folder_items = portal.folder.objectValues('ATDocument')
        folder_items.sort(sortFunc)
        old_version, current_version = folder_items

        ## Check that reference now targets ``new_version``
        self.failUnless(other_doc.getRelatedItems() == [new_version])


    def testNewVersion(self):
        """
        Test 'StagingTool.{get,has,isThe}NewVersion()'
        """
        portal = self.portal
        stool = getToolByName(portal, STOOL_ID)
        getNewVersion = stool.getNewVersion
        hasNewVersion = stool.hasNewVersion
        isTheNewVersion = stool.isTheNewVersion
        wtool = getToolByName(portal, 'portal_workflow')
        document = portal.folder.document1

        ## No new version
        self.login('manager')
        wtool.doActionFor(document, 'publish')
        self.failUnless(getNewVersion(document) is None)
        self.failUnless(not hasNewVersion(document))
        self.failUnless(not isTheNewVersion(document))

        ## 'manager' can get (i.e. view) the new version
        new_version = stool.createNewVersion(document, redirect=False)
        self.failUnless(getNewVersion(document) == new_version)
        self.failUnless(hasNewVersion(document))
        self.failUnless(not isTheNewVersion(document))
        self.failUnless(getNewVersion(new_version) is None)
        self.failUnless(not hasNewVersion(new_version))
        self.failUnless(isTheNewVersion(new_version))

        ## 'member1' can also get the new version
        self.login('member1')
        self.failUnless(getNewVersion(document) == new_version)
        self.failUnless(hasNewVersion(document))
        self.failUnless(not isTheNewVersion(document))
        self.failUnless(getNewVersion(new_version) is None)
        self.failUnless(not hasNewVersion(new_version))
        self.failUnless(isTheNewVersion(new_version))

        ## Hide the new version: 'manager' should still be able to get
        ## it...
        self.login('manager')
        wtool.doActionFor(new_version, 'hide')
        self.failUnless(getNewVersion(document) == new_version)
        self.failUnless(hasNewVersion(document))
        self.failUnless(not isTheNewVersion(document))
        self.failUnless(getNewVersion(new_version) is None)
        self.failUnless(not hasNewVersion(new_version))
        self.failUnless(isTheNewVersion(new_version))

        ## ... but 'member1' should not be able to get it.
        self.login('member1')
        self.failUnless(getNewVersion(document) is None)
        self.failUnless(hasNewVersion(document))

        ## Remove the new version
        self.login('manager')
        document.aq_parent.manage_delObjects([new_version.getId()])
        self.failUnless(getNewVersion(document) is None)
        self.failUnless(not hasNewVersion(document))
        self.failUnless(getattr(document, '_new_version', None) is None)
        self.login('member1')
        self.failUnless(getNewVersion(document) is None)
        self.failUnless(not hasNewVersion(document))

        ## Create a new version and "promote" (i.e. publish) it.
        self.login('manager')
        new_version = stool.createNewVersion(document, redirect=False)
        wtool.doActionFor(new_version, 'publish')
        new_version = portal.folder.document1
        old_version = document

        self.failUnless(getNewVersion(new_version) is None)
        self.failUnless(not hasNewVersion(new_version))
        self.failUnless(not isTheNewVersion(new_version))
        self.failUnless(getNewVersion(old_version) is None)
        self.failUnless(not hasNewVersion(old_version))
        self.failUnless(not isTheNewVersion(old_version))


    def testGetAndIsCurrentVersion(self):
        """
        Test 'StagingTool.{get,is}CurrentVersion()'
        """
        portal = self.portal
        stool = getToolByName(portal, STOOL_ID)
        getCurrentVersion = stool.getCurrentVersion
        isCurrentVersion = stool.isCurrentVersion
        wtool = getToolByName(portal, 'portal_workflow')
        document = portal.folder.document1

        self.login('manager')
        wtool.doActionFor(document, 'publish')
        self.failUnless(getCurrentVersion(document) ==  document)
        self.failUnless(isCurrentVersion(document))

        ## Create a new version
        new_version = stool.createNewVersion(document, redirect=False)
        self.failUnless(getCurrentVersion(document) ==  document)
        self.failUnless(isCurrentVersion(document))
        self.failUnless(getCurrentVersion(new_version) ==  document)
        self.failUnless(not isCurrentVersion(new_version))

        ## Hide 'document' so that a member cannot get it
        wtool.doActionFor(document, 'retract')
        wtool.doActionFor(document, 'hide')
        self.login('member1')
        self.failUnless(getCurrentVersion(document) is None)
        self.failUnless(getCurrentVersion(new_version) is None)
        self.failUnless(isCurrentVersion(document))

        ## Promote 'new_version'.
        self.login('manager')
        current_version, old_version = stool.promoteVersion(new_version)
        document = new_version = None ## Just to be sure that we do
                                      ## not mix variable names
        self.failUnless(getCurrentVersion(current_version) == current_version)
        self.failUnless(isCurrentVersion(current_version))
        self.failUnless(getCurrentVersion(old_version) == current_version)
        self.failUnless(not isCurrentVersion(old_version))


    def testGetVersions(self):
        """
        Test 'StagingTool.getVersions()'
        """
        portal = self.portal
        stool = getToolByName(portal, STOOL_ID)
        stool['plone_workflow'].setExpirePreviouslyPublishedVersion(True)
        wtool = getToolByName(portal, 'portal_workflow')
        folder = portal.folder
        document = folder.document1
        sortFunc = lambda x, y: cmp(x.creation_date,
                                    y.creation_date)
        getVersions = lambda x: stool.getVersions(x, sortFunc)

        ## Hide the item from 'member1'
        self.login('manager')
        wtool.doActionFor(document, 'hide')
        self.failUnless(getVersions(document) ==  [document])
        self.login('member1')
        self.failUnless(getVersions(document) ==  [])

        ## Publish it so that 'member1' can view it
        self.login('manager')
        wtool.doActionFor(document, 'show')
        wtool.doActionFor(document, 'publish')
        self.failUnless(getVersions(document) ==  [document])
        self.login('member1')
        self.failUnless(getVersions(document) ==  [document])

        ## Create a new version
        self.login('manager')
        new_version = stool.createNewVersion(document, redirect=False)
        self.failUnless(getVersions(document) == [document, new_version])
        self.failUnless(getVersions(new_version) == [document, new_version])
        self.login('member1')
        self.failUnless(getVersions(document) == [document, new_version])
        self.failUnless(getVersions(new_version) == [document, new_version])

        ## Hide 'new_version' so that a member cannot get it
        self.login('manager')
        wtool.doActionFor(new_version, 'hide')
        self.failUnless(getVersions(document) == [document, new_version])
        self.failUnless(getVersions(new_version) == [document, new_version])
        self.login('member1')
        self.failUnless(getVersions(document) == [document])
        self.failUnless(getVersions(new_version) == [document])

        ## Also hide 'document', for the fun
        self.login('manager')
        wtool.doActionFor(document, 'retract')
        wtool.doActionFor(document, 'hide')
        self.failUnless(getVersions(document) == [document, new_version])
        self.failUnless(getVersions(new_version) == [document, new_version])
        self.login('member1')
        self.failUnless(getVersions(document) == [])
        self.failUnless(getVersions(new_version) == [])

        ## Show 'document' again and promote 'new_version'
        self.login('manager')
        wtool.doActionFor(document, 'show')
        wtool.doActionFor(new_version, 'show')
        current_version, old_version = stool.promoteVersion(new_version)
        document = new_version = None ## Just to be sure that we do
                                      ## not mix variable names
        self.failUnless(getVersions(current_version) == [old_version,
                                                         current_version])
        self.failUnless(getVersions(old_version) == [old_version,
                                                     current_version])
        self.login('member1')
        self.failUnless(getVersions(current_version) == [old_version,
                                                         current_version])
        self.failUnless(getVersions(old_version) == [old_version,
                                                     current_version])

        ## Remove 'old_version'
        self.login('manager')
        folder.manage_delObjects([old_version.getId()])
        self.failUnless(getVersions(current_version) == [current_version])
        self.login('member1')
        self.failUnless(getVersions(current_version) == [current_version])


    def testIsAnOldVersion(self):
        """
        Test 'StagingTool.isAnOldVersion()'
        """
        portal = self.portal
        stool = getToolByName(portal, STOOL_ID)
        isAnOldVersion = stool.isAnOldVersion
        wtool = getToolByName(portal, 'portal_workflow')
        document = portal.folder.document1

        self.login('manager')
        wtool.doActionFor(document, 'publish')
        self.failUnless(not isAnOldVersion(document))

        ## Create a new version
        new_version = stool.createNewVersion(document, redirect=False)
        self.failUnless(not isAnOldVersion(document))
        self.failUnless(not isAnOldVersion(new_version))

        ## Promote 'new_version'.
        self.login('manager')
        current_version, old_version = stool.promoteVersion(new_version)
        document = new_version = None ## Just to be sure that we do
                                      ## not mix variable names
        self.failUnless(not isAnOldVersion(current_version))
        self.failUnless(isAnOldVersion(old_version))


    def itemsAreEquivalent(self, orig, copy, ignore_fields):
        """
        Test that 'orig' and 'copy' have the same schema values

        Helper method for 'testDuplicationOfSimpleItems()' and
        'testDuplicationOfFolderishItems()'.
        """
        copy_schema = copy.Schema()
        for orig_field in orig.Schema().fields():
            if orig_field.getName() in ignore_fields:
                continue
            orig_value = orig_field.get(orig)
            copy_field = copy_schema.getField(orig_field.getName())
            copy_value = copy_field.get(copy)
            self.failUnless(orig_value == copy_value,
                            "Values for field '%s' of items '%s' "
                            "and '%s' are not equal: "
                            "%s -> %s" % (orig_field.getName(),
                                          orig.absolute_url(1),
                                          copy.absolute_url(2),
                                          orig_value,
                                          copy_value))


    def foldersAreEquivalent(self, orig, copy, ignore_fields):
        """
        Test that 'orig' and 'copy' are equivalent (see above) and
        that their subitems are equivalent (recursively)

        Helper method for 'testDuplicationOfFolderishItems()'.
        """
        self.itemsAreEquivalent(orig, copy, ignore_fields)
        for orig_subitem in orig.objectValues():
            copy_subitem = copy[orig_subitem.getId()]

            testFunc = self.itemsAreEquivalent
            if orig_subitem.is_folderish():
                testFunc = self.foldersAreEquivalent
            testFunc(orig_subitem, copy_subitem, ignore_fields)


    def testDuplicationOfSimpleItems(self):
        """
        Test 'StagingTool._duplicateObject()' on a simple (non-
        folderish) item
        """
        portal = self.portal
        stool = getToolByName(portal, STOOL_ID)
        isAnOldVersion = stool.isAnOldVersion
        wtool = getToolByName(portal, 'portal_workflow')
        folder = portal.folder
        document = folder.document1
        self.login('manager')
        document.setRelatedItems([folder.UID()])
        document.reindexObject(['related_items'])
        NEW_ID = 'new_id'
        IGNORE_FIELDS = ('id', 'creation_date', 'modification_date')

        orig = document
        copy = stool._duplicateObject(orig, NEW_ID)
        self.itemsAreEquivalent(orig, copy, IGNORE_FIELDS)


    def testDuplicationOfFolderishItems(self):
        """
        Test 'StagingTool._duplicateObject()' on a folderish item
        """
        portal = self.portal
        stool = getToolByName(portal, STOOL_ID)
        isAnOldVersion = stool.isAnOldVersion
        wtool = getToolByName(portal, 'portal_workflow')
        folder = portal.folder
        document = folder.document1
        self.login('manager')
        document.setRelatedItems([folder.UID()])
        document.reindexObject(['related_items'])
        NEW_ID = 'new_id'
        IGNORE_FIELDS = ('id', 'creation_date', 'modification_date')

        orig = folder
        copy = stool._duplicateObject(orig, NEW_ID)
        self.foldersAreEquivalent(orig, copy, IGNORE_FIELDS)


    ## The starting '_' is intentional. See below.
    def _testLinguaPloneSupport(self):
        """
        Test LinguaPlone support
        """
        portal = self.portal
        stool = getToolByName(portal, STOOL_ID)
        qitool = getToolByName(portal, 'portal_quickinstaller')
        wtool = getToolByName(portal, 'portal_workflow')
        folder = portal.folder
        document = folder.document1

        ## Install LinguaPlone and add another supported language
        self.login('manager')
        qitool.installProduct('LinguaPlone')
        qitool.installProduct('PloneLanguageTool')
        ltool = getToolByName(portal, 'portal_languages')
        ltool.addSupportedLanguage('fr')

        ## Translate a document in French
        document.setLanguage('en')
        document.addTranslation('fr')
        fr_translation = document.getTranslation('fr')

        ## This is the "normal" behaviour of LinguaPlone items
        self.failUnless(document.isCanonical())
        self.failUnless(document.getTranslation('fr') == fr_translation)
        self.failUnless(fr_translation.getTranslation('en') == document)
        self.failUnless(fr_translation.getCanonical() == document)

        ## Create a new version of the canonical document
        wtool.doActionFor(document, 'publish')
        new_version = stool.createNewVersion(document, redirect=False)
        new_version, old_version = stool.promoteVersion(new_version)

        ## Tests
        self.failUnless(new_version.isCanonical())
        translations = [t[0] for t in old_version.getTranslations().values()]
        self.failUnless(translations == [old_version], translations)
        self.failUnless(fr_translation.getTranslation('en') == new_version)
        self.failUnless(new_version.getTranslation('fr') == fr_translation)
        self.failUnless(fr_translation.getCanonical() == new_version)

        ## And create a new version of the French translation
        wtool.doActionFor(fr_translation, 'publish')
        new_fr_translation = stool.createNewVersion(fr_translation,
                                                    redirect=False)

        ## Tests
        new_fr_translation, old_fr_translation = stool.promoteVersion(new_fr_translation)
        translations = [t[0] for t in old_fr_translation.getTranslations().values()]
        self.failUnless(translations == [old_fr_translation], translations)
        self.failUnless(new_fr_translation.getTranslation('en') == new_version)
        self.failUnless(new_version.getTranslation('fr') == new_fr_translation)
        self.failUnless(new_fr_translation.getCanonical() == new_version)

    if HAS_LINGUA_PLONE:
        ## We obviously do not want to test LinguaPlone support if
        ## LinguaPlone is not installed.
        testLinguaPloneSupport = _testLinguaPloneSupport


    def testStagingInformationsAreIgnoredWhenCopyPaste(self):
        """
        Test that staging informations are ignored (i.e. removed) when
        an item is copied and pasted
        """
        portal = self.portal
        stool = getToolByName(portal, STOOL_ID)
        wtool = getToolByName(portal, 'portal_workflow')
        folder = portal.folder
        document = folder.document1

        self.login('manager')
        ## Set staging informations by creating a new version
        wtool.doActionFor(document, 'publish')
        new_version = stool.createNewVersion(document, redirect=False)

        ## Copy and paste the new version
        cp = folder.manage_copyObjects(new_version.getId())
        result = folder.manage_pasteObjects(cp)
        copy = folder[result[0]['new_id']]
        self.failUnless(stool.getVersions(copy) == [copy])


    def testStagingInformationsAreKeptdWhenCutPaste(self):
        """
        Test that staging informations are kept when an item is cut
        and pasted
        """
        portal = self.portal
        stool = getToolByName(portal, STOOL_ID)
        wtool = getToolByName(portal, 'portal_workflow')
        folder = portal.folder
        subfolder = folder.subfolder
        document = folder.document1

        self.login('manager')
        ## Set staging informations by creating a new version
        wtool.doActionFor(document, 'publish')
        new_version = stool.createNewVersion(document, redirect=False)

        ## Cut and paste the new version
        cp = folder.manage_cutObjects(new_version.getId())
        result = subfolder.manage_pasteObjects(cp)
        pasted = subfolder[result[0]['new_id']]
        sortFunc = lambda x, y: cmp(x.creation_date,
                                    y.creation_date)
        versions = stool.getVersions(pasted, sortFunc)
        self.failUnless(versions == [document, pasted])


    def testRemoveWorkflowHistory(self):
        """
        Test 'StagingTool._removeWorkflowHistory()'
        """
        portal = self.portal
        stool = getToolByName(portal, STOOL_ID)
        wtool = getToolByName(portal, 'portal_workflow')
        folder = portal.folder
        document = folder.document1

        self.login('manager')
        wtool.doActionFor(document, 'publish')
        new_version = stool.createNewVersion(document, redirect=False)

        ## Default initial state for 'plone_workflow' is 'visible'
        state = wtool.getInfoFor(new_version, 'review_state')
        self.failUnless(state == 'visible')

        ## Test that history has been reset
        history = wtool.getInfoFor(new_version, 'review_history')
        self.failUnless(len(history) == 1)

        ## Set initial state to 'private'...
        wf = wtool['plone_workflow'].states.setInitialState('private')

        ## ... and check that the workflow state of the new version is
        ## the initial state of the workflow.
        folder.manage_delObjects([new_version.getId()])
        new_version = stool.createNewVersion(document, redirect=False)
        state = wtool.getInfoFor(new_version, 'review_state')
        self.failUnless(state == 'private', state)


    def testStagingKeepOrdering(self):
        """Test that staging features keep items ordered as they should.

        """
        portal = self.portal
        stool = getToolByName(portal, STOOL_ID)
        wtool = getToolByName(portal, 'portal_workflow')
        folder = portal.folder
        getOrderedItems = lambda: folder.getFolderContents(full_objects=True)

        self.login('manager')

        ## We want our own dummy items.
        folder.invokeFactory('Document', id='foo')
        foo = folder['foo']
        wtool.doActionFor(foo, 'publish')
        folder.invokeFactory('Document', id='bar')
        bar = folder['bar']
        wtool.doActionFor(bar, 'publish')
        folder.invokeFactory('Document', id='baz')
        baz = folder['baz']
        wtool.doActionFor(baz, 'publish')

        ## ... therefore remove standard dummy ones.
        folder.manage_delObjects(['document1', 'subfolder'])

        ## A new version should be put just after the initial position
        ## of the current version.
        self.failUnless(getOrderedItems() == [foo, bar, baz])

        ## And the ordering must be kept when the new version is
        ## published (and therefore replaces the old one).
        new_foo = stool.createNewVersion(foo, redirect=False)
        wtool.doActionFor(new_foo, 'publish')
        self.failUnless(getOrderedItems() == [foo, new_foo, bar, baz])


if  __name__ == '__main__':
    framework()

def test_suite():
    from unittest import TestSuite, makeSuite
    suite = TestSuite()
    suite.addTest(makeSuite(TestStaging))
    return suite
